classdef Node < handle

    properties (Constant, GetAccess = private)
        InputNode = "InputNode"
        DerivedNode = "DerivedNode"
        OutputNode = "OutputNode"
        SubNode = "SubNode"
    end

    properties
        name
    end

    properties (SetAccess = private, GetAccess = public)
        value
        type
        prenodes
        postnodes
        rule_prenodes
        rule_postnodes
        func
        nout
        subnodeind
        flag_initialized
        flag_active
    end

    methods (Static = true)
        function obj = output(name, func, varargin) 
            obj = comg.Node();
            obj.type = obj.OutputNode;
            if isstring(name) || ischar(name)
                obj.name = string(name);
                obj.func = func;
                obj.prenodes = [varargin{:}];
            else
                obj.func = name;
                obj.prenodes = [func, varargin{:}];
            end
            for nd = obj.prenodes
                nd.postnodes = [nd.postnodes, obj];
            end
            obj.realUpdate();
        end

        function obj = derived(name, nout, func, varargin)
            obj = comg.Node();
            obj.type = obj.DerivedNode;
            obj.nout = 1;
            nds = [varargin{:}];
            
            if isstring(name) || ischar(name)
                obj.name = string(name);
            else
                nds = [func, nds];
                func = nout;
                nout = name;
            end

            if isnumeric(nout)
                obj.nout = nout;
            else
                nds = [func, nds];
                func = nout;
            end

            if obj.nout > 1
                obj.value = cell(1, obj.nout);
            end
            obj.func = func;
            obj.prenodes = nds;
            for nd = obj.prenodes
                nd.postnodes = [nd.postnodes, obj];
            end
            obj.realUpdate();
        end
    
    end

    methods
        function obj = Node(name, value)
            obj.type = obj.InputNode;
            obj.flag_initialized = false;
            obj.flag_active = false;

            if (nargin == 1) && ~isempty(name)
                obj.value = name;
            elseif nargin == 2
                obj.name = string(name);
                if ~isempty(value)
                    obj.value = value;
                end
            end
            
        end

        function obj = activate(obj)
            if ~obj.flag_active
                obj.flag_active = true;
                for nd = obj.prenodes
                    nd.activate();
                end
                obj.realUpdate();
            end
        end

        function obj = after(obj, nd)
            obj.rule_prenodes = [obj.rule_prenodes, nd];
            nd.rule_postnodes = [nd.rule_postnodes, obj];
        end

        function obj = before(obj, nd)
            nd.after(obj);
        end

        function updater = updateValue(obj, value, updater)
            if obj.type ~= obj.InputNode
                error("You can only update the values of input nodes.")
            end

            if ~exist('updater', 'var')
                updater = [];
            end

            obj.value = value;

            flag_lazy = (nargout >= 1);
            if flag_lazy
                updater = obj.update(updater);
            else
                obj.update(updater);
            end
            
        end

        function updater = update(obj, updater)
            flag_lazy = (nargout >= 1);
            if ~exist('updater', 'var')
                updater = [];
            end
            
            updater = obj.preUpdate(updater);

            if ~flag_lazy
                for nd = updater
                    nd.realUpdate();
                end
                updater = [];
            end
        end
        
        function newnode = subnode(obj, name, ind)
            if obj.type ~= obj.DerivedNode
                error("You can only genrate subnode from derived node.");
            end

            newnode = comg.Node();
            newnode.type = obj.SubNode;
            newnode.prenodes = obj;
            
            if isstring(name) || ischar(name)
                newnode.name = string(name);
            else
                ind = name;
            end

            newnode.subnodeind = ind;
            obj.postnodes = [obj.postnodes, newnode];
        end

    end

    methods
        function set.value(obj, value)
            obj.value = value;
            obj.flag_initialized = true; %#ok<MCSUP> 
        end

        function value = get.value(obj)
            if obj.type == obj.SubNode
                value = obj.prenodes.value{obj.subnodeind};
            else
                value = obj.value;
            end
        end

        function flag = get.flag_initialized(obj)
            if obj.type == obj.SubNode
                flag = obj.prenodes.flag_initialized;
            else
                flag = obj.flag_initialized;
            end
        end
    
    end

    methods (Access = private)
        function updater = preUpdate(obj, updater)
            if ~ismember(obj, updater) && obj.flag_active
                for nd = obj.postnodes
                    updater = nd.preUpdate(updater);
                end
                for nd = obj.rule_postnodes
                    updater = nd.preUpdate(updater);
                end
                updater = [obj, updater];
            end
        end

        function realUpdate(obj)
%             fprintf("The updating node is %s, name %s.\n", obj.type, obj.name);
            if obj.type == obj.DerivedNode
                obj.deriveValue();
            elseif obj.type == obj.OutputNode
                obj.showValue();
            end
        end

        function deriveValue(obj)
            if all([obj.prenodes.flag_initialized])
                if obj.nout == 1
                    obj.value = obj.func(obj.prenodes.value);
                else
                    [obj.value{:}] = obj.func(obj.prenodes.value);
                end
            end
        end

        function showValue(obj)
            if all([obj.prenodes.flag_initialized])
                obj.func(obj.prenodes.value);
            end
        end

    end

end

