function [coeff, coef_cent] = tempmatch(template, dst, mask)
% TEMPMATCH function implement template matching algorithm.
% 
% COEFF = tempmatch(TEMPLATE, DST)
%   calculates the Cross-Correlation Coefficients of TEMPLATE and DST. DST
%   is a N-d array with shape (L1,L2,...,LN), TEMPLATE is also a N-d array
%   with shape (K1,K2,...,KN) satisfied Ki <= Li. The output COEFF is the
%   cross-correlation coefficients of TEMPLATE and patches of DST with the
%   same shape at different location. The shape of COEFF is
%   (L1+1-K1,...,LN+1-KN). Specifically,
%
%   COEFF(i1,i2,...,iN) = cc(TEMPLATE, DST(i1:i1+K1-1,...,iN:iN+KN-1))
%
%   where cc(A,B) = <A,B>/(|A||B|) \in [-1,1]. When |A| or |B| is 0,
%   cc(A,B) will be forcedly set as 0.
%
% COEFF = tempmatch(TEMPLATE, DST, MASK)
%   implements the template matching algorithm with mask. The MASK should
%   have the same size as TEMPLATE, when the template slides in the DST,
%   only those pixels in MASK will be count to calculate the cc. MASK can
%   also be viewed as the weight of each pixesl when calculate the inner
%   product and the norm. so MASK can contain non-negtive values besides 0
%   and 1.

% Check the size of template and dst
tempsz = size(template);
if ~exist('mask', 'var')
    mask = ones(tempsz);
end
assert(all(size(mask) == tempsz));
dstsz = size(dst);
assert(numel(dstsz) >= numel(tempsz));
tempsz = [tempsz, ones(numel(dstsz)-numel(tempsz), 1)];
assert(all(dstsz >= tempsz));
coefsz = dstsz - tempsz + 1;
eps = 1e-10;

% remove the mean
tempmean = sum(template.*mask, 'all')./sum(mask, 'all');
template = template - tempmean;
temppow = sum(template.^2.*mask, 'all');
if (temppow < eps)
    coeff = zeros(coefsz);
    return;
end

% calculate the move-inner-product
movinnprod = calcmovinnprod(template.*mask, dst);

% calculate the mov-power
movsum = calcmovinnprod(mask, dst);
movpow = calcmovinnprod(mask, dst.^2);
movpow = movpow - (movsum.^2)./sum(mask,'all');

% calculate coefficients
coeff = movinnprod ./ sqrt(movpow * temppow);
coeff(movpow < 1e-10) = 0;

coef_cent = cenpos(size(dst)) - cenpos(size(template)) + 1;

end


function movinnprod = calcmovinnprod(template, dst)
    tempsz = size(template);
    dstsz = size(dst);
    tempsz = [tempsz, ones(numel(dstsz)-numel(tempsz), 1)];
    coefsz = dstsz - tempsz + 1;
    Ndim = numel(dstsz);

    padtemp = zeros(dstsz);
    cmd = 'padtemp(';
    for idim = 1:Ndim
        cmd = [cmd, '1:tempsz(', num2str(idim), '),'];
    end
    cmd = [cmd(1:end-1), ') = template;'];
    eval(cmd);
    
    fttemp = fftn(padtemp);
    ftdst = fftn(dst);
    movinnprod = real(ifftn(ftdst.*conj(fttemp)));
    cmd = 'movinnprod = movinnprod(';
    for idim = 1:Ndim
        cmd = [cmd, '1:coefsz(', num2str(idim), '),'];
    end
    cmd = [cmd(1:end-1), ');'];
    eval(cmd);
end