classdef TomoFilamentViewer < handle

    properties (SetAccess = private, GetAccess = private)
        handle_tomo_layer
        handle_selecting_filament
        handle_selecting_filament_pts
        handle_all_filaments
        handle_all_filaments_pts
        handle_filament_slider
        handle_filament_section
        handle_refine_point_in_section
        handle_refine_center_tag_in_section
        handle_creating_filament

        tomo_layer_alpha
        selecting_filament_line_color
        all_filaments_line_color
        filament_slider_alpha
        filament_slider_radius
        filament_slider_single_side_length
        filament_slider_current_length % in curve

        old_wind_butt_moti_fcn
        old_wind_butt_down_fcn
        old_wind_butt_up_fcn

        in_filament_edit_mode
        in_filament_create_mode
        editing_filament_rela_coords
        editing_filament_rela_curve

        creating_filament_coords
    end

    properties
        default_view_angle
        
        tomo_axes
        section_axes
        layer_slider
        filament_list

        tomo_vol
        filament_data

        filament_sample_step
        filament_extract_radius
        filament_section_averaging_length

        current_layer_idx
        current_filament_idx
        current_filament_smoothed_points
        current_filament_curve
        current_filament_anchors_length % the length at curve
        current_filament_full_length
        current_filament_R
        current_filament_S
        current_filament_extracted

        current_refine_point_idx

    end

    methods
        function obj = TomoFilamentViewer(tomo_axes, layer_slider, filament_list, section_axes)
            obj.filament_sample_step = 1;
            obj.filament_extract_radius = 30;

            obj.default_view_angle = [-80, 70];
            obj.selecting_filament_line_color = [152, 245, 255] / 255;
            obj.all_filaments_line_color = [100, 200, 100] / 255;
            obj.filament_slider_radius = 15;
            obj.filament_slider_single_side_length = 20;
            obj.filament_section_averaging_length = 1;
            obj.tomo_layer_alpha = 0.9;
            obj.filament_slider_alpha = 0.8;

            obj.in_filament_edit_mode = false;
            obj.in_filament_create_mode = false;

            obj.tomo_axes = tomo_axes;
            obj.layer_slider = layer_slider;
            obj.filament_list = filament_list;
            obj.section_axes = section_axes;

            obj.initAxes();
            obj.initSlider();
            obj.initListBox();
        end

        function initAxes(obj)
            title(obj.tomo_axes, "Tomo data view");
            xlabel(obj.tomo_axes, "X-aixs (fastest axis in tomo data)");
            ylabel(obj.tomo_axes, "Y-aixs");
            zlabel(obj.tomo_axes, "Z-aixs");
            colormap(obj.tomo_axes, 'gray');
            obj.tomo_axes.View = obj.default_view_angle;
            hold(obj.tomo_axes, "on");
            axis(obj.tomo_axes, "equal");

            title(obj.section_axes, "Section view");
            colormap(obj.section_axes, 'gray');
            hold(obj.section_axes, "on");
            axis(obj.section_axes, "equal");
        end

        function initSlider(obj)
            obj.layer_slider.Enable = false;
        end

        function initListBox(obj)
            obj.filament_list.Enable = false;
        end

        function loadTomoData(obj, tomo_vol)
            obj.tomo_vol = tomo_vol;
            obj.tomo_vol = (obj.tomo_vol - min(obj.tomo_vol,[],'all'))/(max(obj.tomo_vol,[],'all') - min(obj.tomo_vol,[],'all'));
            xlim(obj.tomo_axes, 0.5+[0, size(obj.tomo_vol,1)]);
            ylim(obj.tomo_axes, 0.5+[0, size(obj.tomo_vol,2)]);
            zlim(obj.tomo_axes, 0.5+[0, size(obj.tomo_vol,3)]);
            obj.current_layer_idx = cenpos(size(obj.tomo_vol,3));

            obj.layer_slider.Limits = [1, size(obj.tomo_vol, 3)];
            obj.layer_slider.Value = obj.current_layer_idx;
            obj.layer_slider.ValueChangingFcn = @(src, evt)(obj.sliderValueChangingCallback(evt));
            obj.layer_slider.ValueChangedFcn = @(src, evt)(obj.sliderValueChangedCallback(evt));
            obj.layer_slider.Enable = true;

            obj.updateAxesContents();
        end

        function loadFilamentData(obj, filament_data)
            obj.filament_list.ValueChangedFcn = @(src, evt)(obj.listValueChangedCallback(evt));
            obj.filament_data = filament_data;

            obj.updateFilamentList();
            obj.filament_list.Enable = true;
        end

        function setCurrentLayer(obj, idx)
            obj.current_layer_idx = idx;

            obj.updateAxesContents();
        end

        function updateAxesContents(obj)
            xImage = 0.5+[0, size(obj.tomo_vol,1); 0, size(obj.tomo_vol,1)];   % The x data for the image corners
            yImage = 0.5+[0, 0; size(obj.tomo_vol,2), size(obj.tomo_vol,2)];   % The y data for the image corners
            zImage = obj.current_layer_idx * [1, 1; 1, 1];             % The z data for the image corners

            img = obj.tomo_vol(:,:,obj.current_layer_idx)';

            if ~isempty(obj.handle_tomo_layer)
                delete(obj.handle_tomo_layer);
            end
            obj.handle_tomo_layer = surf(obj.tomo_axes, xImage,yImage,zImage,...    % Plot the surface
                 'CData',img * 255, ...
                 'FaceColor', 'texturemap',...
                 'EdgeColor', 'none', ...
                 'ButtonDownFcn', @(src, evt)(0));
            alpha(obj.handle_tomo_layer, obj.tomo_layer_alpha);
        end

        function drawFilamentSliderAt(obj, ptidx)
            if ~exist('ptidx','var')
                ptidx = obj.filament_slider_current_length;
            end

            if ptidx < 1
                ptidx = 1;
            elseif ptidx > obj.current_filament_full_length
                ptidx = obj.current_filament_full_length;
            end
            obj.filament_slider_current_length = ptidx;

            start_idx = max(1, round(ptidx - obj.filament_slider_single_side_length));
            end_idx = min(round(obj.current_filament_full_length), round(ptidx + obj.filament_slider_single_side_length));

            theta = 0:5:360;
            xx = obj.current_filament_smoothed_points(1, start_idx:end_idx)' + ...
                obj.current_filament_R(1, start_idx:end_idx)' .* cosd(theta) * obj.filament_slider_radius + ...
                obj.current_filament_S(1, start_idx:end_idx)' .* sind(theta) * obj.filament_slider_radius;
            yy = obj.current_filament_smoothed_points(2, start_idx:end_idx)' + ...
                obj.current_filament_R(2, start_idx:end_idx)' .* cosd(theta) * obj.filament_slider_radius + ...
                obj.current_filament_S(2, start_idx:end_idx)' .* sind(theta) * obj.filament_slider_radius;
            zz = obj.current_filament_smoothed_points(3, start_idx:end_idx)' + ...
                obj.current_filament_R(3, start_idx:end_idx)' .* cosd(theta) * obj.filament_slider_radius + ...
                obj.current_filament_S(3, start_idx:end_idx)' .* sind(theta) * obj.filament_slider_radius;

            if ~isempty(obj.handle_filament_slider)
                delete(obj.handle_filament_slider); 
            end
            obj.handle_filament_slider = surf(obj.tomo_axes, ...
                xx,yy,zz, ...
                'FaceColor', '#80B3FF',...
                'EdgeColor', 'none', 'LineStyle', 'none', ...
                'ButtonDownFcn', @(src, evt)obj.filamentSliderButtonDownCallback(evt));
            alpha(obj.handle_filament_slider, obj.filament_slider_alpha);

            if ~isempty(obj.handle_filament_section)
                delete(obj.handle_filament_section);
            end
            obj.handle_filament_section = imagesc(obj.section_axes, ...
                -obj.filament_extract_radius:obj.filament_extract_radius, ...
                -obj.filament_extract_radius:obj.filament_extract_radius, ...
                mean(obj.current_filament_extracted(:,:,start_idx:end_idx), 3)');

            if obj.in_filament_edit_mode
                center_pos = fnval(obj.editing_filament_rela_curve, obj.filament_slider_current_length);
                delete(obj.handle_refine_center_tag_in_section);
                obj.handle_refine_center_tag_in_section = plot(obj.section_axes, ...
                    center_pos(1), center_pos(2), 'rx', ...
                    'MarkerSize', 20, ...
                    'ButtonDownFcn', @(src, evt)(0));

                delete(obj.handle_refine_point_in_section);
                [anchor_dis, anchor_idx] = min(abs(obj.editing_filament_rela_coords(3,:) - ptidx));
                if anchor_dis < obj.filament_slider_single_side_length
                    obj.handle_refine_point_in_section = ...
                        drawpoint(obj.section_axes, ...
                        'Position', obj.editing_filament_rela_coords(1:2,anchor_idx)');
                    addlistener(obj.handle_refine_point_in_section,'ROIMoved',@(src, evt)obj.refinementPointMovedCallback(evt));
                    obj.current_refine_point_idx = anchor_idx;
                else
                    obj.current_refine_point_idx = [];
                end

            end

        end

        function updateFilamentCurve(obj)
            if isempty(obj.filament_list.Value)
                delete(obj.handle_selecting_filament);
                delete(obj.handle_selecting_filament_pts);
                delete(obj.handle_filament_slider);
                delete(obj.handle_filament_section);
                delete(obj.handle_refine_point_in_section);
                delete(obj.handle_refine_center_tag_in_section);
                return
            end

            selecting_vals = split(obj.filament_list.Value);
            obj.current_filament_idx = str2double(selecting_vals(2));
            anchor_points = obj.filament_data.data{obj.current_filament_idx}.Points;

            [~, obj.current_filament_curve, obj.current_filament_full_length, obj.current_filament_anchors_length] = ...
                obj.filamentSmoothing(anchor_points);
            
            [obj.current_filament_extracted,~,~,~,obj.current_filament_R,obj.current_filament_S,~,obj.current_filament_smoothed_points] = ...
                filamentExtraction(obj.tomo_vol, obj.current_filament_curve, obj.filament_sample_step, obj.current_filament_full_length, obj.filament_extract_radius);

            delete(obj.handle_selecting_filament);
            obj.handle_selecting_filament = plot3(obj.tomo_axes, obj.current_filament_smoothed_points(1,:), ...
                obj.current_filament_smoothed_points(2,:), obj.current_filament_smoothed_points(3,:), ...
                'Color', obj.selecting_filament_line_color, ...
                'LineWidth', 2);

            delete(obj.handle_selecting_filament_pts);
            if obj.in_filament_edit_mode
                editing_anchor_len = obj.editing_filament_rela_coords(3,:);
                draw_points = fnval(obj.current_filament_curve, editing_anchor_len);
            else
                draw_points = anchor_points;
            end
            obj.handle_selecting_filament_pts = plot3(obj.tomo_axes, draw_points(1,:), ...
               draw_points(2,:), draw_points(3,:), ...
                'o', 'Color', obj.selecting_filament_line_color, ...
                'LineWidth', 2, ...
                'ButtonDownFcn', @(src, evt)(0));
            if obj.in_filament_edit_mode
                obj.drawFilamentSliderAt(obj.filament_slider_current_length);
            else
                obj.drawFilamentSliderAt(1);
            end
        end

        function updateFilamentList(obj)
            if obj.filament_data.current_index <= 0
                obj.filament_list.Items = {};
            else
                obj.filament_list.Items = arrayfun(@(id)("Filament " + id), 1:obj.filament_data.current_index);
            end
        end

        function createFilament(obj)
            obj.in_filament_create_mode = true;
            obj.old_wind_butt_down_fcn = obj.tomo_axes.Parent.WindowButtonDownFcn;
            obj.old_wind_butt_moti_fcn = obj.tomo_axes.Parent.WindowButtonMotionFcn;
            obj.creating_filament_coords = [];
            obj.tomo_axes.Parent.WindowButtonDownFcn = @(src, evt)(obj.creatingFilamentClickCallback(evt));
            obj.tomo_axes.Parent.WindowButtonMotionFcn = @(src, evt)(obj.creatingFilamentMotionCallback(evt));

            obj.filament_list.Value = {};
            obj.updateFilamentCurve();

        end

        function finishCreatingFilament(obj)
            obj.in_filament_create_mode = false;
            obj.tomo_axes.Parent.WindowButtonDownFcn = obj.old_wind_butt_down_fcn;
            obj.tomo_axes.Parent.WindowButtonMotionFcn = obj.old_wind_butt_moti_fcn;

            if ~isempty(obj.handle_creating_filament)
                delete(obj.handle_creating_filament);
            end

            if size(obj.creating_filament_coords, 2) > 0
                obj.filament_data.addFilament(obj.creating_filament_coords);
                obj.creating_filament_coords = [];
                obj.updateFilamentList();
            end

            obj.filament_list.Value = obj.filament_list.Items(end);
            obj.updateFilamentCurve();
        end

        function enterFilamentEditMode(obj)
            obj.filament_list.Enable = false;
            obj.in_filament_edit_mode = true;

            Npt = length(obj.current_filament_anchors_length);
            obj.editing_filament_rela_coords = [zeros(2,Npt);obj.current_filament_anchors_length];
            obj.editing_filament_rela_curve = csapi(obj.editing_filament_rela_coords(3,:), obj.editing_filament_rela_coords);

            obj.drawFilamentSliderAt();

        end

        function exitFilamentEditMode(obj)
            obj.filament_list.Enable = true;
            obj.in_filament_edit_mode = false;
            if ~isempty(obj.handle_selecting_filament_pts)
                delete(obj.handle_selecting_filament_pts);
            end

            [~, ord] = sort(obj.editing_filament_rela_coords(3,:));
            rela_cds = obj.editing_filament_rela_coords(:,ord);
            rela_idx_f = 1 + floor(rela_cds(3,:));
            rela_idx_c = 1 + ceil(rela_cds(3,:));
            abs_cds_f = obj.current_filament_smoothed_points(:, rela_idx_f) + ...
                obj.current_filament_R(:, rela_idx_f).*rela_cds(1, :) + ...
                obj.current_filament_S(:, rela_idx_f).*rela_cds(2, :);
            abs_cds_c = obj.current_filament_smoothed_points(:, rela_idx_c) + ...
                obj.current_filament_R(:, rela_idx_c).*rela_cds(1, :) + ...
                obj.current_filament_S(:, rela_idx_c).*rela_cds(2, :);

            abs_cds = abs_cds_f + (abs_cds_c-abs_cds_f).*(rela_idx_c-rela_idx_f);

            obj.editing_filament_rela_coords = [];

            obj.filament_data.modifyFilament(obj.current_filament_idx, abs_cds);
            obj.updateFilamentCurve();
            
        end

        function showAllFilaments(obj)
            obj.hideAllFilaments();
            Nfila = obj.filament_data.current_index;
            obj.handle_all_filaments = cell(Nfila, 1);
            for ifila = 1:Nfila
                obj.handle_all_filaments{ifila} = plot3(obj.tomo_axes, ...
                    obj.filament_data.data{ifila}.Points(1,:), ...
                    obj.filament_data.data{ifila}.Points(2,:), ...
                    obj.filament_data.data{ifila}.Points(3,:), ...
                    'Color', obj.all_filaments_line_color, ...
                    'LineWidth', 2);
            end
            
        end

        function hideAllFilaments(obj)
            if ~isempty(obj.handle_all_filaments)
                for ihandle = 1:numel(obj.handle_all_filaments)
                    delete(obj.handle_all_filaments{ihandle});
                end
            end
        end

        function moveToNextAnchor(obj)
            next_anchor_idx = find(obj.editing_filament_rela_coords(3,:) >= obj.filament_slider_current_length + 1, 1);
            if ~isempty(next_anchor_idx)
                obj.drawFilamentSliderAt(round(obj.editing_filament_rela_coords(3,next_anchor_idx)));
            end
        end

        function moveToPreviousAnchor(obj)
            next_anchor_idx = find(obj.editing_filament_rela_coords(3,:) <= obj.filament_slider_current_length - 1, 1, 'last');
            if ~isempty(next_anchor_idx)
                obj.drawFilamentSliderAt(round(obj.editing_filament_rela_coords(3,next_anchor_idx)));
            end
        end

        function addAnchorAtCurrentSection(obj)
            if obj.in_filament_edit_mode && (isempty(obj.handle_refine_point_in_section) || ~isvalid(obj.handle_refine_point_in_section))
                obj.editing_filament_rela_coords = [obj.editing_filament_rela_coords, [0;0;obj.filament_slider_current_length]];
                [~, ord] = sort(obj.editing_filament_rela_coords(3,:));
                obj.editing_filament_rela_coords = obj.editing_filament_rela_coords(:,ord);
                obj.editing_filament_rela_curve = csapi(obj.editing_filament_rela_coords(3,:), obj.editing_filament_rela_coords);
                obj.updateFilamentCurve();
                obj.drawFilamentSliderAt(obj.filament_slider_current_length);
            end
        end

        function deleteAnchorAtCurrentSection(obj)
            if obj.in_filament_edit_mode && (~isempty(obj.handle_refine_point_in_section) && isvalid(obj.handle_refine_point_in_section))
                obj.editing_filament_rela_coords(:, obj.current_refine_point_idx) = [];
                obj.editing_filament_rela_curve = csapi(obj.editing_filament_rela_coords(3,:), obj.editing_filament_rela_coords);
                obj.updateFilamentCurve();
                obj.drawFilamentSliderAt(obj.filament_slider_current_length);
            end
        end

    end

    
    methods (Static = true, Access = public) % some utility function
        function [smoothed_points, filament_curve, filament_length, filament_points_length] = filamentSmoothing(points)
            filament_points_length = [0,cumsum(sqrt(sum(diff(points,[],2).^2,1)))];
            filament_curve = csapi(filament_points_length, points);
            filament_length = filament_points_length(end);
            smoothed_points = fnval(filament_curve, 1:filament_length);
        end
    end



    methods (Access = private) % call-back functions
        function sliderValueChangingCallback(obj, evt)
            obj.current_layer_idx = round(evt.Value);
            obj.updateAxesContents();
        end

        function sliderValueChangedCallback(obj, evt)
            obj.current_layer_idx = round(evt.Value);
            obj.updateAxesContents();
        end

        function listValueChangedCallback(obj, ~)
            obj.updateFilamentCurve();
        end

        function filamentSliderButtonDownCallback(obj, ~)
            fig = obj.tomo_axes.Parent;

            obj.old_wind_butt_moti_fcn = fig.WindowButtonMotionFcn;
            fig.WindowButtonMotionFcn = @(src, evt)obj.filamentSliderDraggingCallback(evt);

            obj.old_wind_butt_up_fcn = fig.WindowButtonUpFcn;
            fig.WindowButtonUpFcn = @(src, evt)obj.filamentSliderButtonUpCallback(evt);
        end

        function filamentSliderDraggingCallback(obj, ~)
            pt = obj.tomo_axes.CurrentPoint';
            
            filpts = obj.current_filament_smoothed_points;

            diff = filpts - pt(:,1);
            dire = normalize(pt(:,2) - pt(:,1), 'norm');
            innp = sum(dire .* diff, 1);
            proj = innp .* dire;
            perp = diff - proj;
            dist = sqrt(sum(perp.^2, 1));

            [~, new_pos_idx] = min(dist);
            obj.drawFilamentSliderAt(new_pos_idx);
        end

        function filamentSliderButtonUpCallback(obj, ~)
            fig = obj.tomo_axes.Parent;
            fig.WindowButtonMotionFcn = obj.old_wind_butt_moti_fcn;
            fig.WindowButtonUpFcn = obj.old_wind_butt_up_fcn;
        end

        function creatingFilamentMotionCallback(obj, ~)
            pt = obj.tomo_axes.CurrentPoint;
            pt_at_current_layer = pt(1,:) + (pt(2,:) - pt(1,:)) / (pt(2,3)-pt(1,3)) * (obj.current_layer_idx-pt(1,3));
            cds = [obj.creating_filament_coords, pt_at_current_layer'];

            if ~isempty(obj.handle_creating_filament)
                delete(obj.handle_creating_filament);
            end
            obj.handle_creating_filament = plot3(obj.tomo_axes, cds(1,:), cds(2,:), cds(3,:), ...
                '-ro', 'LineWidth', 2, 'ButtonDownFcn', @(src, evt)(0));
        end

        function creatingFilamentClickCallback(obj, ~)
            if obj.tomo_axes.Parent.SelectionType == "alt"
                if (size(obj.creating_filament_coords, 2) > 1)
                    obj.creating_filament_coords = obj.creating_filament_coords(:, 1:end-1);
                    obj.creatingFilamentMotionCallback();
                end
                return
            end

            pt = obj.tomo_axes.CurrentPoint;
            pt_at_current_layer = pt(1,:) + (pt(2,:) - pt(1,:)) / (pt(2,3)-pt(1,3)) * (obj.current_layer_idx-pt(1,3));

            cond = all(pt_at_current_layer >= 1) && all(pt_at_current_layer <= size(obj.tomo_vol)) && ...
                all(pt_at_current_layer <= [obj.tomo_axes.XLim(2),obj.tomo_axes.YLim(2),obj.tomo_axes.ZLim(2)]) && ...
                all(pt_at_current_layer >= [obj.tomo_axes.XLim(1),obj.tomo_axes.YLim(1),obj.tomo_axes.ZLim(1)]);
            if cond
                if ~isempty(obj.creating_filament_coords) && (norm(obj.creating_filament_coords(:,end) - pt_at_current_layer') < 1)
                    obj.finishCreatingFilament();
                    return
                end
                obj.creating_filament_coords = [obj.creating_filament_coords, pt_at_current_layer'];
                obj.creatingFilamentMotionCallback();
            end
        end

        function refinementPointMovedCallback(obj, evt)
            obj.editing_filament_rela_coords(1:2,obj.current_refine_point_idx) = evt.CurrentPosition;
            obj.editing_filament_rela_curve = csapi(obj.editing_filament_rela_coords(3,:), obj.editing_filament_rela_coords);
            obj.drawFilamentSliderAt();
        end

    end


end